from threading import Lock, Thread

from .icalparser import parse_events
from .icaldownload import ICalDownload


# Lock for event data
event_lock = Lock()
# Event data
event_store = {}
# Threads
threads = {}


def events(url=None, file=None, string_content=None, start=None, end=None, fix_apple=False):
    """
    Get all events form the given iCal URL occurring in the given time range.

    :param url: iCal URL
    :param file: iCal file path
    :param string_content: iCal content as string
    :param start: start date (see dateutils.date)
    :param end: end date (see dateutils.date)
    :param fix_apple: fix known Apple iCal issues
    :return: events as list of dictionaries
    """
    found_events = []

    content = None

    if url:
        content = ICalDownload().data_from_url(url, apple_fix=fix_apple)

    if not content and file:
        content = ICalDownload().data_from_file(file, apple_fix=fix_apple)

    if not content and string_content:
        content = ICalDownload().data_from_string(string_content,
                                                  apple_fix=fix_apple)

    found_events += parse_events(content, start=start, end=end)

    return found_events


def request_data(key, url, file, string_content, start, end, fix_apple):
    """
    Request data, update local data cache and remove this Thread form queue.

    :param key: key for data source to get result later
    :param url: iCal URL
    :param file: iCal file path
    :param string_content: iCal content as string
    :param start: start date
    :param end: end date
    :param fix_apple: fix known Apple iCal issues
    """
    data = []

    try:
        data += events(url=url, file=file, string_content=string_content,
                       start=start, end=end, fix_apple=fix_apple)
    finally:
        update_events(key, data)
        request_finished(key)


def events_async(key, url=None, file=None, start=None, string_content=None,
                 end=None, fix_apple=False):
    """
    Trigger an asynchronous data request.

    :param key: key for data source to get result later
    :param url: iCal URL
    :param file: iCal file path
    :param string_content: iCal content as string
    :param start: start date
    :param end: end date
    :param fix_apple: fix known Apple iCal issues
    """
    t = Thread(target=request_data, args=(key, url, file, string_content, start, end, fix_apple))

    with event_lock:
        if key not in threads:
            threads[key] = []

        threads[key].append(t)

        if not threads[key][0].is_alive():
            threads[key][0].start()


def request_finished(key):
    """
    Remove finished Thread from queue.

    :param key: data source key
    """
    with event_lock:
        threads[key] = threads[key][1:]

        if threads[key]:
            threads[key][0].run()


def update_events(key, data):
    """
    Set the latest events for a key.

    :param key: key to set
    :param data: events for key
    """
    with event_lock:
        event_store[key] = data


def latest_events(key):
    """
    Get the latest downloaded events for the given key.

    :return: events for key
    """
    with event_lock:
        # copy data
        res = event_store[key][:]

    return res


def all_done(key):
    """
    Check if requests for the given key are active.

    :param key: key for requests
    :return: True if requests are pending or active
    """
    with event_lock:
        if threads[key]:
            return False
        return True



import unittest
from icalevents import icalevents
from datetime import date
from time import sleep


class ICalEventsTests(unittest.TestCase):
    def test_events_url(self):
        url = "https://raw.githubusercontent.com/irgangla/icalevents/master/test/test_data/basic.ics"
        start = date(2017, 5, 18)
        end = date(2017, 5, 19)

        evs = icalevents.events(url=url, file=None, start=start, end=end)

        self.assertEqual(len(evs), 2, "two events are found")

    def test_events_start(self):
        ical = "test/test_data/basic.ics"
        start = date(2017, 5, 16)

        evs = icalevents.events(url=None, file=ical, start=start)

        self.assertEqual(len(evs), 3, "three events are found")

    def test_events(self):
        ical = "test/test_data/basic.ics"
        start = date(2017, 5, 18)
        end = date(2017, 5, 19)

        evs = icalevents.events(url=None, file=ical, start=start, end=end)

        self.assertEqual(len(evs), 2, "two events are found")

    def test_event_attributes(self):
        ical = "test/test_data/basic.ics"
        start = date(2017, 7, 12)
        end = date(2017, 7, 13)

        ev = icalevents.events(url=None, file=ical, start=start, end=end)[0]

        self.assertEqual(ev.summary, "graue Restmülltonne")
        self.assertEqual(ev.description, "graue Restmülltonne nicht vergessen!")
        self.assertTrue(ev.all_day)

    def test_events_async_url(self):
        url = "https://raw.githubusercontent.com/irgangla/icalevents/master/test/test_data/basic.ics"
        start = date(2017, 5, 18)
        end = date(2017, 5, 19)
        key = "basic"

        icalevents.events_async(key, url=url, file=None, start=start, end=end)

        sleep(4)

        self.assertTrue(icalevents.all_done(key), "request is finished")
        self.assertEqual(len(icalevents.latest_events(key)), 2, "two events are found")

    def test_events_async(self):
        ical = "test/test_data/basic.ics"
        start = date(2017, 5, 18)
        end = date(2017, 5, 19)
        key = "basic"

        icalevents.events_async(key, url=None, file=ical, start=start, end=end)

        sleep(4)

        self.assertTrue(icalevents.all_done(key), "request is finished")
        self.assertEqual(len(icalevents.latest_events(key)), 2, "two events are found")

    def test_request_data(self):
        ical = "test/test_data/basic.ics"
        start = date(2017, 5, 18)
        end = date(2017, 5, 19)
        key = "basic"

        icalevents.request_data(key, url=None, file=ical, string_content=None, start=start, end=end, fix_apple=False)

        self.assertTrue(icalevents.all_done(key), "request is finished")
        self.assertEqual(len(icalevents.latest_events(key)), 2, "two events are found")

    def test_string_data(self):
        ical = "test/test_data/basic.ics"

        with open(ical, mode='rb') as f:
            string_content = f.read()

        start = date(2017, 5, 18)
        end = date(2017, 5, 19)
        key = "basic"

        icalevents.request_data(key, url=None, file=None, string_content=string_content, start=start, end=end,
                                fix_apple=False)

        self.assertTrue(icalevents.all_done(key), "request is finished")
        self.assertEqual(len(icalevents.latest_events(key)), 2, "two events are found")



"""
Parse iCal data to Events.
"""
# for UID generation
from random import randint
from datetime import datetime, timedelta, date
from dateutil import relativedelta

from icalendar import Calendar
from icalendar.prop import vDDDLists
from pytz import utc


# default query length (one week)
default_span = timedelta(days=7)


def now():
    """
    Get current time.

    :return: now as datetime with timezone
    """
    return utc.localize(datetime.utcnow())


class Event:
    """
    Represents one event (occurrence in case of reoccurring events).
    """

    def __init__(self):
        """
        Create a new event occurrence.
        """
        self.uid = -1
        self.summary = None
        self.description = None
        self.start = None
        self.end = None
        self.all_day = True

    def time_left(self, time=now()):
        """
        timedelta form now to event.

        :return: timedelta from now
        """
        return self.start - time

    def __lt__(self, other):
        """
        Events are sorted by start time by default.

        :param other: other event
        :return: True if start of this event is smaller than other
        """
        if not other or not isinstance(other, Event):
            raise ValueError('Only events can be compared with each other! Other is %s' % type(other))
        else:
            return self.start < other.start

    def __str__(self):
        n = now()

        # compute time delta description
        if not self.all_day:
            if self.end > n > self.start:
                # event is now
                delta = "now"
            elif self.start > n:
                # event is a future event
                if self.time_left().days > 0:
                    delta = "%s days left" % self.time_left().days
                else:
                    hours = self.time_left().seconds / (60 * 60)
                    delta = "%.1f hours left" % hours
            else:
                # event is over
                delta = "ended"
        else:
            if self.end > n > self.start:
                delta = "today"
            elif self.start > n:
                delta = "%s days left" % self.time_left().days
            else:
                delta = "ended"

        return "%s: %s (%s)" % (self.start, self.summary, delta)

    def copy_to(self, new_start=None, uid=None):
        """
        Create a new event equal to this with new start date.

        :param new_start: new start date
        :param uid: UID of new event
        :return: new event
        """
        duration = self.end - self.start

        if not new_start:
            new_start = self.start

        if not uid:
            uid = "%s_%d" % (self.uid, randint(0, 1000000))

        ne = Event()
        ne.summary = self.summary
        ne.description = self.description
        ne.start = new_start
        ne.end = (new_start + duration)
        ne.all_day = (self.all_day and (new_start - self.start).seconds == 0)
        ne.uid = uid

        return ne


def next_year_at(dt, count=1):
    """
    Move date <count> years to the future.

    :param dt: date as datetime
    :param count: number of years
    :return: date datetime
    """
    dt += relativedelta.relativedelta(years=+count)
    return normalize(datetime(year=dt.year, month=dt.month, day=dt.day,
                              hour=dt.hour, minute=dt.minute,
                              second=dt.second, microsecond=dt.microsecond))


def next_month_at(dt, count=1):
    """
    Move date <count> months to the future.

    :param dt: date as datetime
    :param count: number of months
    :return: date datetime
    """
    dt += relativedelta.relativedelta(months=+count)

    return normalize(datetime(year=dt.year, month=dt.month, day=dt.day,
                              hour=dt.hour, minute=dt.minute, second=dt.second,
                              microsecond=dt.microsecond))


def create_event(component):
    """
    Create an event from its iCal representation.

    :param component: iCal component
    :return: event
    """

    event = Event()

    event.start = normalize(component.get('dtstart').dt)
    
    if component.get('dtend'):
        event.end = normalize(component.get('dtend').dt)
    elif component.get('duration'):
        event.end = event.start + component.get('duration').dt
    else:
        raise ValueError("Event has neither end, nor duration property.")
    
    event.summary = str(component.get('summary'))
    event.description = str(component.get('description'))
    event.all_day = type(component.get('dtstart').dt) is date

    return event


def normalize(dt):
    """
    Convert date or datetime to datetime with timezone.

    :param dt: date to normalize
    :return: date as datetime with timezone
    """
    if type(dt) is date:
        dt = datetime(dt.year, dt.month, dt.day, 0, 0)
    elif type(dt) is datetime:
        pass
    else:
        raise ValueError("unknown type %s" % type(dt))

    if not dt.tzinfo:
        dt = utc.localize(dt)

    return dt


def find_last(rule, freq, end, first):
    last = end

    if rule.get('UNTIL'):
        last = rule.get('UNTIL')[0]
    elif rule.get('COUNT'):
        count = rule.get('COUNT')[0]
        if freq == 'YEARLY':
            last = next_year_at(first.start, count=count)
        elif freq == 'MONTHLY':
            last = next_month_at(first.start, count=count)
        elif freq == 'WEEKLY':
            last = normalize(first.start + timedelta(days=7*count))
        elif freq == 'DAILY':
            last = normalize(first.start + timedelta(days=count))

    return normalize(last)


def in_range(event_list, start, end):
    """
    Find elements in the given time range.

    :param event_list: list of events
    :param start: start datetime
    :param end: end datetime
    :return: events in range
    """
    filtered = []
    for e in event_list:
        if e.end > start and e.start < end:
            filtered.append(e)

    return filtered


def parse_events(content, start=None, end=None):
    """
    Query the events occurring in a given time range.

    :param content: iCal URL/file content as String
    :param start: start date for search, default today
    :param end: end date for search
    :return: events as list
    """
    if not start:
        start = now()

    if not end:
        end = start + default_span

    if not content:
        raise ValueError('Content is invalid!')

    calendar = Calendar.from_ical(content)

    start = normalize(start)
    end = normalize(end)

    found = []

    for component in calendar.walk():
        if component.name == "VEVENT":
            if component.get('rrule'):
                es = create_recurring_events(start, end, component)
                if es:
                    found += es
            else:
                e = create_event(component)
                if e.end >= start and e.start <= end:
                    found.append(e)
    return found


def create_recurring_events(start, end, component):
    """
    Unfold a reoccurring event to its occurrances into the given time frame.

    :param start: start of the time frame
    :param end: end of the time frame
    :param component: iCal component
    :return: occurrances of the event
    """
    start = normalize(start)
    end = normalize(end)

    rule = component.get('rrule')

    unfolded = []

    first = create_event(component)
    unfolded.append(first)

    freq = str(rule.get('FREQ')[0])
    last = find_last(rule, freq, end, first)

    if last < start:
        return
    elif end < last:
        limit = end
    else:
        limit = last

    current = first

    if freq == 'YEARLY':
        while True:
            current = current.copy_to(next_year_at(current.start))
            if current.start < limit:
                unfolded.append(current)
            else:
                break
    elif freq == 'MONTHLY':
        by_day = rule.get('BYDAY')

        while True:
            if by_day:
                next_date = next_month_byday_delta(current.start, by_day[0])
                current = current.copy_to(next_date)
            else:
                current = current.copy_to(next_month_at(current.start))
            if current.start < limit:
                unfolded.append(current)
            else:
                break
    elif freq == 'DAILY':
        delta = timedelta(days=1)
        while True:
            current = current.copy_to(current.start + delta)
            if current.start < limit:
                unfolded.append(current)
            else:
                break
    elif freq == 'WEEKLY':
        delta = timedelta(days=7)

        by_day = rule.get('BYDAY')
        if by_day:
            day_deltas = generate_day_deltas_by_weekday(set(by_day))
        else:
            day_deltas = None

        while True:
            if day_deltas:
                delta = timedelta(days=day_deltas.get(current.start.weekday()))
            current = current.copy_to(current.start + delta)
            if current.start < limit:
                unfolded.append(current)
            else:
                break
    else:
        return

    reduced_range = in_range(unfolded, start, end)

    exceptions = extract_exdates(component)
    reduced_exceptions = [ev for ev in reduced_range if ev.start not in exceptions]

    return reduced_exceptions


def extract_exdates(component):
    """
    Compile a list of all exception dates stored with a component.
    
    :param component: icalendar iCal component
    :return: exception dates
    """
    dates = []

    exd_prop = component.get('exdate')
    if exd_prop:
        if isinstance(exd_prop, list):
            for exd_list in exd_prop:
                dates.extend(exd.dt for exd in exd_list.dts)
        elif isinstance(exd_prop, vDDDLists):
            dates.extend(exd.dt for exd in exd_prop.dts)

    return dates


def generate_day_deltas_by_weekday(by_day):
    """
    Create a dict to determine how many days to add to the current
    event to get the next event when a WEEKLY rule contains a
    BYDAY clause.

    The resulting dictionary has the weekday number as keys and the
    number of days to add to get the next event as values. Weekday
    numbers are the same as those assigned by the date.weekday()
    function.

    :param by_day: list or set of two-letter weekday abbreviations
    :return: dict mapping weekday numbers to day deltas
    """
    weekdays = ['MO', 'TU', 'WE', 'TH', 'FR', 'SA', 'SU']

    selected_weekday_numbers = []
    day_deltas = []
    hop_count = 0
    for weekday_number, weekday_name in enumerate(weekdays):
        if weekday_name in by_day:
            selected_weekday_numbers.append(weekday_number)
            day_deltas.append(hop_count)
            hop_count = 0
        hop_count += 1

    # readjust the first deltas to include the remaining deltas
    first_hop_count = day_deltas[0] + hop_count
    adjusted_deltas = day_deltas[1:] + [first_hop_count]

    return dict(zip(selected_weekday_numbers, adjusted_deltas))


def next_month_byday_delta(start_date, by_day):
    """
    Get the next event date when a MONTHLY rule contains a BYDAY clause,
    e.g. 3SA = "Next third Saturday"
    """

    weekdays = ['MO', 'TU', 'WE', 'TH', 'FR', 'SA', 'SU']

    number = int(by_day[0])
    weekday = by_day[1:]

    if weekday not in weekdays:
        raise ValueError('Invalid weekday: {}'.format(weekday))

    weekday = weekdays.index(weekday)
    weekday_func = relativedelta.weekday(weekday)

    delta = relativedelta.relativedelta(day=1, months=+1,
                                        weekday=weekday_func(number))

    return start_date + delta



"""
Downloads an iCal url or reads an iCal file.
"""
from httplib2 import Http


# default http connection to use
default_http = Http('.cache')


def apple_data_fix(content):
    """
    Fix Apple tzdata bug.

    :param content: content to fix
    :return: fixed content
    """
    return content.replace('TZOFFSETFROM:+5328', 'TZOFFSETFROM:+0053')


def apple_url_fix(url):
    """
    Fix Apple URL.

    :param url: URL to fix
    :return: fixed URL
    """
    if url.startswith("webcal://"):
        url = url.replace('webcal://', 'http://', 1)
    return url


class ICalDownload:
    """
    Downloads or reads and decodes iCal sources.
    """
    def __init__(self, http=default_http, encoding='utf-8'):
        self.http = http
        self.encoding = encoding

    def data_from_url(self, url, apple_fix=False):
        """
        Download iCal data from URL.

        :param url: URL to download
        :param apple_fix: fix Apple bugs (protocol type and tzdata in iCal)
        :return: decoded (and fixed) iCal data
        """
        if apple_fix:
            url = apple_url_fix(url)

        _, content = self.http.request(url)

        if not content:
            raise ConnectionError('Could not get data from %s!' % url)

        return self.decode(content, apple_fix=apple_fix)

    def data_from_file(self, file, apple_fix=False):
        """
        Read iCal data from file.

        :param file: file to read
        :param apple_fix: fix wrong Apple tzdata in iCal
        :return: decoded (and fixed) iCal data
        """
        with open(file, mode='rb') as f:
            content = f.read()

        if not content:
            raise IOError("File %f is not readable or is empty!" % file)

        return self.decode(content, apple_fix=apple_fix)

    def data_from_string(self, string_content, apple_fix=False):
        if not string_content:
            raise IOError("String content is not readable or is empty!")

        return self.decode(string_content, apple_fix=apple_fix)

    def decode(self, content, apple_fix=False):
        """
        Decode content using the set charset.

        :param content: content do decode
        :param apple_fix: fix Apple txdata bug
        :return: decoded (and fixed) content
        """
        content = content.decode(self.encoding)
        content = content.replace('\r', '')

        if apple_fix:
            content = apple_data_fix(content)

        return content
